"""
Support functions for JADN codec
  Convert dict between nested and flat
  Convert typedef options between dict and strings
"""

from functools import reduce
from jadn.definitions import *

# Dict conversion utilities


def _dmerge(x, y):
    k, v = next(iter(y.items()))
    if k in x:
        _dmerge(x[k], v)
    else:
        x[k] = v
    return x


def hdict(keys, value, sep='.'):
    """
    Convert a hierarchical-key value pair to a nested dict
    """
    return reduce(lambda v, k: {k: v}, reversed(keys.split(sep)), value)


def fluff(src, sep='.'):
    """
    Convert a flat dict with hierarchical keys to a nested dict

    :param src: flat dict - e.g., {'a.b.c': 1, 'a.b.d': 2}
    :param sep: separator character for keys
    :return: nested dict - e.g., {'a': {'b': {'c': 1, 'd': 2}}}
    """
    return reduce(lambda x, y: _dmerge(x, y), [hdict(k, v, sep) for k, v in src.items()], {})


def flatten(cmd, path='', fc=None, sep='.'):
    """
    Convert a nested dict to a flat dict with hierarchical keys
    """
    if fc is None:
        fc = {}
    fcmd = fc.copy()
    if isinstance(cmd, dict):
        for k, v in cmd.items():
            k = k.split(':')[1] if ':' in k else k
            fcmd = flatten(v, sep.join((path, k)) if path else k, fcmd)
    elif isinstance(cmd, list):
        for n, v in enumerate(cmd):
            fcmd.update(flatten(v, sep.join([path, str(n)])))
    else:
        fcmd[path] = cmd
    return fcmd


def dlist(src):
    """
    Convert dicts with numeric keys to lists

    :param src: {'a': {'b': {'0':'red', '1':'blue'}, 'c': 'foo'}}
    :return: {'a': {'b': ['red', 'blue'], 'c': 'foo'}}
    """
    if isinstance(src, dict):
        for k in src:
            src[k] = dlist(src[k])
        if set(src) == set([str(k) for k in range(len(src))]):
            src = [src[str(k)] for k in range(len(src))]
    return src


def multiplicity(minc, maxc):
    if minc == 1 and maxc == 1:
        return '1'
    return str(minc) + '..' + ('*' if maxc == 0 else str(maxc))


def topo_sort(items):
    """
    Topological sort with locality
    Sorts a list of (item: (dependencies)) pairs so that 1) all dependency items are listed before the parent item,
    and 2) dependencies are listed in the given order and as close to the parent as possible.
    Returns the sorted list of items and a list of root items.  A single root indicates a fully-connected hierarchy;
    multiple roots indicate disconnected items or hierarchies, and no roots indicate a dependency cycle.
    """
    def walk_tree(it):
        for i in deps[it]:
            if i not in out:
                walk_tree(i)
                out.append(i)

    out = []
    deps = {i[0]: i[1] for i in items}           # TODO: update for items dict
    roots = {i[0] for i in items} - set().union(*[i[1] for i in items])
    for item in roots:
        walk_tree(item)
        out.append(item)
    out = out if out else [i[0] for i in items]     # if cycle detected, don't sort
    return out, roots


def topts_s2d(olist):        # Convert list of type definition option strings to options dictionary
    tval = {
        'id': lambda x: True,
        'vtype': lambda x: x,
        'ktype': lambda x: x,
        'enum': lambda x: x,
        'pointer': lambda x: x,
        'format': lambda x: x,
        'pattern': lambda x: x,
        'minv': lambda x: int(x),
        'maxv': lambda x: int(x),
        'unique': lambda x: True,
        'and': lambda x: x,
        'or': lambda x: x,
    }

    assert set(tval) == {k for k in TYPE_OPTIONS.values()}
    assert isinstance(olist, (list, tuple)), '%r is not a list' % olist
    opts = {}
    for o in olist:
        try:
            k = TYPE_OPTIONS[ord(o[0])]
            opts[k] = tval[k](o[1:])
        except KeyError:
            raise ValueError('Unknown type option: %s' % o)
    return opts


def ftopts_s2d(ostr):       # Convert list of field definition option strings to options dictionary
    fval = {
        'minc': lambda x: int(x),
        'maxc': lambda x: int(x),
        'tfield': lambda x: x,
        'dir': lambda x: True,
        'default': lambda x: x,
    }

    assert set(fval) == {k for k in FIELD_OPTIONS.values()}
    assert isinstance(ostr, (list, tuple)), '%r is not a list' % ostr
    fopts, topts = {}, {}
    for o in ostr:
        try:
            k = FIELD_OPTIONS[ord(o[0])]
            fopts[k] = fval[k](o[1:])
        except KeyError:
            topts.update(topts_s2d([o]))
    return fopts, topts


def typestring(typename, typeopts):            # Convert typename and options to string.

    def _typestr(tname, opts):          # SIDE EFFECT: remove known options from opts to flag leftovers

        def _kvstr(optv):               # Handle ktype/vtype containing Enum options
            if optv[0] == OPTION_ID['enum']:
                return 'Enum(' + optv[1:] + ')'
            elif optv[0] == OPTION_ID['pointer']:
                return 'Pointer(' + optv[1:] + ')'
            return optv

        def _rangestr(ops, vrange=False):  # Value range is double-ended: [*..*], Size range is single-ended: [0..*]
            dlo = '*' if vrange else 0
            lo = ops.pop('minv', dlo)
            hi = ops.pop('maxv', '*')
            hi = '*' if hi == 0 and not vrange else hi
            return str(lo) + '..' + str(hi) if lo != dlo or hi != '*' else ''

        extra = '.ID' if opts.pop('id', None) else ''   # SIDE EFFECT!: remove known options from opts.
        if tname == 'ArrayOf':
            extra += '(' + _kvstr(opts.pop('vtype')) + ')'
        elif tname == 'MapOf':
            extra += '(' + _kvstr(opts.pop('ktype')) + ', ' + _kvstr(opts.pop('vtype')) + ')'
        v = opts.pop('enum', None)
        extra += '(Enum(' + v + '))' if v else ''
        v = opts.pop('pointer', None)
        extra += '(Pointer(' + v + '))' if v else ''
        v = opts.pop('pattern', None)
        extra += '(%' + v + '%)' if v else ''
        v = _rangestr(opts, tname in ('Integer', 'Number'))
        extra += '{' + v + '}' if v else ''
        v = opts.pop('format', None)
        extra += ' /' + v if v else ''
        v = opts.pop('unique', None)
        extra += ' unique' if v else ''
        v = opts.pop('and', None)           # hack set operations for now.  TODO: generalize to any number
        extra += ' ∩ ' + v if v else ''
        v = opts.pop('or', None)
        extra += ' ∪ ' + v if v else ''
        return tname + extra

    o2 = typeopts.copy()                        # Don't modify opts
    ts = _typestr(typename, o2)
    return ts + (' ?' + str([str(k) for k in o2]) + '?' if o2 else '')  # Flag unrecognized options


def get_config(meta):
    config = dict(DEFAULT_CONFIG)
    config.update(meta['config'] if meta and 'config' in meta else {})
    return config



import copy
import os
from collections import defaultdict
from jadn.definitions import *
from jadn.core import SchemaModule


def resolve_imports(schema, dirname, no_nsid=()):       # Add referenced types to schema. dirname => other schema files

    def merge_tname(tref, module, imports, nsids):
        """
        Convert reference to an imported type (nsid:TypeName) to a local type. Return unchanged if local.
        :param tref: Type name to be merged into the base module
        :param module: Namespace (unique name) of base module
        :param imports: Dict that maps base module's namespace ids to namespaces
        :param nsids: Dict that maps each namespace to a namespace id. If blank, do not append $nsid qualifier
        :return: Local type name, qualified (TypeName$nsid) or unqualified (TypeName)
        """

        sys = '$'  # Character reserved for use in tool-generated type names
        nsid, tname = tref.split(':', maxsplit=1) if ':' in tref else ('', tref)
        ns_id = nsids[imports[nsid] if nsid else module][0]
        return tname + sys + ns_id if ns_id and not is_builtin(tname) else tname

    def merge_typedef(tdef, module, imports, nsids):
        def update_opts(opts):
            return [x[0] + merge_tname(x[1:], module, imports, nsids) if x[0] in oids else x for x in opts]

        oids = [OPTION_ID['ktype'], OPTION_ID['vtype'], OPTION_ID['and']]  # Options whose value is/has a type name
        tn = merge_tname(tdef[TypeName], module, imports, nsids)
        td = [tn, tdef[BaseType], tdef[TypeOptions], tdef[TypeDesc]]
        td[TypeOptions] = update_opts(td[TypeOptions])
        if len(tdef) > Fields:
            new_fields = copy.deepcopy(tdef[Fields])
            if td[BaseType] != 'Enumerated':
                for f in new_fields:
                    f[FieldOptions] = update_opts(f[FieldOptions])
                    f[FieldType] = merge_tname(f[FieldType], module, imports, nsids)
            td.append(new_fields)
        return td

    def make_enum(sm, tname):
        if tname[0] in (OPTION_ID['enum'], OPTION_ID['pointer']):
            sys = '$'
            tn = tname[1:]
            if tn not in sm.used and tn in sm.tx:
                if tname[0] == OPTION_ID['enum']:
                    print('  Make enum', tname)
                    edef = [tn + sys + 'Enum', 'Enumerated', [], '', [[f[0], f[1], ''] for f in sm.tx[tn][Fields]]]
                else:
                    print('  Make pointers', tname)
                    edef = [tn + sys + 'Point', 'Enumerated', [], '', [[f[0], f[1], ''] for f in sm.tx[tn][Fields]]]
                sm.schema['types'].append(edef)
            return(True)

    def add_types(sm, tname):        # add referenced typenames in this module to used list
        if {tname} - sm.used:
            sm.add_used(tname)
            try:
                [add_types(sm, tn) for tn in sm.deps[tname]]
            except KeyError as e:
                if not make_enum(sm, tname):
                    print('Error:', e, 'not defined in', sm.module, '(' + sm.source + ')')

    def resolve(sm, types):       # add referenced types from other modules to used list
        if set(types) - sm.used:
            sm.load()
            [add_types(sm, tn) for tn in types]
            for mod in sm.refs:
                if mod in modules:
                    resolve(modules[mod], {t for k, v in sm.refs[mod].items() if k in sm.used for t in v})
                    print('  Merge', mod, 'into', sm.module)
                else:
                    print('Error: module', mod, 'not found.')

    # if 'imports' not in schema['meta']:
    #    return schema
    root = SchemaModule(schema)
    modules = {root.module: root}
    nsids = defaultdict(list)
    for fn in (os.path.join(dirname, f) for f in os.listdir(dirname) if os.path.splitext(f)[1] == '.jadn'):
        sm = SchemaModule(fn)
        if sm.module not in modules:            # Add new module to list
            modules.update({sm.module: sm})
        elif root.module == sm.module and root.schema == sm.schema:     # Update source of root schema if found
            modules[sm.module].source = fn
        elif sm.source != fn:                   # Flag multiple files with same module name
            print('* Duplicate module', sm.module, sm.source, 'Ignoring', fn)
        for id, m in sm.imports.items():
            nsids[m].append('' if id in no_nsid else id)
    types = root.schema['meta']['exports'] if 'exports' in root.schema['meta'] else {}
    resolve(root, types)

    for t in root.used.copy():
        if t[0] in (OPTION_ID['enum'], OPTION_ID['pointer']):
            if t[1:] in root.used:
                root.used.remove(t)     # Don't need explicit type if base type is present
            else:
                print('  Error: no base type for', t)


    # Copy all needed types from other modules into root
    nsids[root.module] = ['']
    sc = {'meta': {k: v for k, v in root.schema['meta'].items() if k != 'imports'}, 'types': []}    # Remove imports
    for sm in [root] + [m for m in modules.values() if m.module != root.module]:
        sc['types'] += [merge_typedef(t, sm.module, sm.imports, nsids) for t in sm.schema['types'] if t[TypeName] in sm.used]
    return sc



import copy
from collections import defaultdict
from jadn.definitions import *
from jadn.utils import topts_s2d, ftopts_s2d


def jadn_strip(schema, width=0):             # Strip or truncate comments from schema
    def estrip(s, n):
        return s[:n-2] + (s[n-2:], '..')[len(s) > n] if n > 1 else s[:n]

    sc = copy.deepcopy(schema)
    for tdef in sc['types']:
        tdef[TypeDesc] = estrip(tdef[TypeDesc], width)
        if len(tdef) > Fields:
            fd = ItemDesc if tdef[BaseType] == 'Enumerated' else FieldDesc
            for fdef in tdef[Fields]:
                fdef[fd] = estrip(fdef[fd], width)
    return sc


def jadn_simplify(schema, extensions=JADN_EXTENSIONS):      # Remove schema extensions
    """
    Given an input schema, return a simplified schema with schema extensions removed.

    1) AnonymousType:   Replace all anonymous type definitions with explicit
    2) Multiplicity:    Replace all multiple-value fields with explicit ArrayOf type definitions
    3) DerivedEnum:     Replace all derived enumerations with explicit Enumerated type definitions
    4) MapOfEnum:       Replace all MapOf types with listed keys with explicit Map type definitions
    5) Pointer:         Replace all pointer enumerations with explicit Map {json_pointer: [id_path]}
    """

    def get_optx(opts, oname):
        n = [i for i, x in enumerate(opts) if x[0] == OPTION_ID[oname]]
        return n[0] if n else None

    def del_opt(opts, oname):
        n = [i for i, x in enumerate(opts) if x[0] == OPTION_ID[oname]]
        if n:
            del opts[n[0]]

    def opts_d2s(to):
        return [OPTION_ID[k] + str(v) for k, v in to.items()]

    def epname(topts):
        ex = get_optx(topts, 'enum')
        px = get_optx(topts, 'pointer')
        if ex is not None:
            rtype = topts[ex][1:]
            oname = 'Enum'
        elif px is not None:
            rtype = topts[px][1:]
            oname = 'Pointer'
        else:
            return None
        return rtype + sys + oname + ('-Id' if get_optx(topts, 'id') else '')

    def simplify_anonymous_types():          # Replace anonymous types in fields with explicit type definitions
        new_types = []
        for tdef in tdefs:
            if has_fields(tdef[BaseType]):
                for fdef in tdef[Fields]:
                    fo, fto = ftopts_s2d(fdef[FieldOptions])
                    if fto:                 # If FieldOptions contains a type option, generate an explicit type
                        name = epname(fdef[FieldOptions])       # Use derived enum typename
                        typeopts = []
                        for o in fto:       # Move all type options to new type
                            typeopts.append(fdef[FieldOptions].pop(get_optx(fdef[FieldOptions], o)))
                        newname = name if name else tdef[TypeName] + sys + fdef[FieldName]
                        if newname not in [t[TypeName] for t in new_types]:
                            new_types.append([newname, fdef[FieldType], typeopts, fdef[FieldDesc]])
                        fdef[FieldType] = newname
        return new_types

    def simplify_multiplicity():            # Replace field multiplicity with explicit ArrayOf type definitions
        new_types = []
        for tdef in tdefs:
            if has_fields(tdef[BaseType]):
                for fdef in tdef[Fields]:
                    fo, fto = ftopts_s2d(fdef[FieldOptions])
                    if ('maxc' in fo and fo['maxc'] != 1) or ('minc' in fo and fo['minc'] > 1):
                        newname = tdef[TypeName] + sys + fdef[FieldName]
                        minc = fo['minc'] if 'minc' in fo else 1
                        newopts = {'vtype': fdef[FieldType], 'minv': max(minc, 1)}  # Don't allow empty ArrayOf
                        newopts.update({'maxv': fo['maxc']} if fo['maxc'] > 1 else {})  # maxv defaults to 0
                        new_types.append([newname, 'ArrayOf', opts_d2s(newopts), fdef[FieldDesc]])
                        fdef[FieldType] = newname  # Point existing field to new ArrayOf
                        f = fdef[FieldOptions]
                        del_opt(f, 'maxc')
                        if minc != 0:
                            del_opt(f, 'minc')
        return new_types

    def simplify_derived_enum():             # Generate Enumerated list of fields or JSON Pointers
        def update_eref(enums, opts, optname, new):
            n = get_optx(opts, optname)
            if n is not None:
                name = epname([opts[n][1:]])
                if name:
                    if name in enums:       # Reference existing Enumerated type
                        opts[n] = opts[n][:1] + enums[name]
                    else:                   # Make new Enumerated type
                        make_items = enum_items if opts[n][1:2] == OPTION_ID['enum'] else pointer_items
                        opts[n] = opts[n][:1] + name
                        new.append([name, 'Enumerated', [], '', make_items(name.split(sys)[0])])

        def enum_items(rtype):
            tdef = tdefs[typex[rtype]]
            fields = tdef[Fields] if has_fields(tdef[BaseType]) else []
            return [[f[FieldID], f[FieldName], f[FieldDesc]] for f in fields]

        def pointer_items(rtype):
            def pathnames(rtype, base=''):                   # Walk subfields of referenced type
                tdef = tdefs[typex[rtype]]
                if has_fields(tdef[BaseType]):
                    for f in tdef[Fields]:
                        if OPTION_ID['dir'] in f[FieldOptions]:
                            yield from pathnames(f[FieldType], f[FieldName] + '/')
                        else:
                            yield [base + f[FieldName], f[FieldDesc]]
            return [[n+1] + f for n, f in enumerate(pathnames(rtype))]

        enums = {}
        for tdef in tdefs:              # Replace enum/pointer options in Enumerated types with explicit items
            if tdef[BaseType] == 'Enumerated':
                to = tdef[TypeOptions]
                rname = epname(to)
                if rname:
                    optx = get_optx(to, 'enum')
                    optx = optx if optx is not None else get_optx(to, 'pointer')
                    items = enum_items if to[optx][:1] == OPTION_ID['enum'] else pointer_items
                    tdef.append(items(to[optx][1:]))
                    del to[optx]
                    enums.update({rname: tdef[TypeName]})
        new_types = []                  # Create new Enumerated enum/pointer types if they don't already exist
        for tdef in tdefs:
            if tdef[BaseType] in ('ArrayOf', 'MapOf'):
                update_eref(enums, tdef[TypeOptions], 'vtype', new_types)
                update_eref(enums, tdef[TypeOptions], 'ktype', new_types)
        return new_types

    def simplify_map_of_enum():           # Replace MapOf(enumerated key) with explicit Map
        for n, tdef in enumerate(tdefs):
            to = topts_s2d(tdef[TypeOptions])
            if tdef[BaseType] == 'MapOf' and tdefs[typex[to['ktype']]][BaseType] == 'Enumerated':
                newfields = [[f[ItemID], f[ItemValue], to['vtype'], [], f[ItemDesc]] for f in tdefs[typex[to['ktype']]][Fields]]
                tdefs[n] = [tdef[TypeName], 'Map', [], tdef[TypeDesc], newfields]

    assert extensions - JADN_EXTENSIONS == set()
    sys = '$'                                   # Character reserved for tool-generated TypeNames
    sc = copy.deepcopy(schema)                  # Don't modify original schema
    tdefs = sc['types']
    if 'AnonymousType' in extensions:           # Expand inline definitions into named type definitions
        tdefs += simplify_anonymous_types()
    if 'Multiplicity' in extensions:            # Expand repeated types into ArrayOf defintions
        tdefs += simplify_multiplicity()
    typex = {t[TypeName]: n for n, t in enumerate(tdefs)}       # Build type index
    if 'DerivedEnum' in extensions:             # Generate Enumerated list of fields or JSON Pointers
        types = {t[TypeName]: t for t in tdefs}
        tdefs += simplify_derived_enum()
    if 'MapOfEnum' in extensions:               # Generate explicit Map from MapOf
        simplify_map_of_enum()
    return sc


def get_pointers(tdef, types):
    def ptr(fdef):
        if OPTION_ID['dir'] in [o[0] for o in fdef[FieldOptions]]:
            return [fdef[FieldName] + '/' + p for p in get_pointers(types[fdef[FieldType]], types)]
        else:
            return fdef[FieldName]

    if has_fields(tdef[BaseType]):
        return [ptr(f) for f in tdef[Fields]]
    else:
        return []


def get_enum_items(tdef, topts, types):
    def ptr(fdef):
        if OPTION_ID['dir'] in fdef[FieldOptions]:
            return fdef[FieldName] + "^"
        return fdef[FieldName]

    if 'enum' in topts:
        return types[topts['enum']][Fields]
    elif 'pointer' in topts:
        return [(n, ptr(f), '') for n, f in enumerate(types[topts['pointer']][Fields])]
    return tdef[Fields]



"""
Load, validate, prettyprint, and dump JSON Abstract Encoding Notation (JADN) schemas
"""

import json
import jsonschema
import numbers
import os
from datetime import datetime
from collections import defaultdict
from jadn.definitions import *
from jadn.utils import topts_s2d, ftopts_s2d
from jadn.codec.codec import Codec


def jadn_check(schema):
    """
    Validate JADN schema against JSON schema,
    Validate JADN schema against JADN meta-schema, then
    Perform additional checks on type definitions
    """

    def check_typeopts(type_name, base_type, topts):  # Check for invalid type options and undefined formats
        ro = {k for k in REQUIRED_TYPE_OPTIONS[base_type]} - {k for k in topts}
        if ro:
            _error('Missing type option', type_name + ':', str(ro))
        uo = {k for k in topts} - {k for k in ALLOWED_TYPE_OPTIONS[base_type]}
        if uo:
            _error('Unsupported type option', type_name + ' (' + base_type+ '):', str(uo))
        if 'format' in topts:
            f = topts['format']
            fm = dict(list(FORMAT_VALIDATE.items()) + list(FORMAT_JS_VALIDATE.items()) + list(FORMAT_SERIALIZE.items()))
            if f not in fm or base_type != fm[f]:
                _error('Unsupported format', f, 'in', type_name, base_type)
        if 'enum' in topts and 'pointer' in topts:
            _error('Type cannot be both Enum and Pointer', type_name, base_type)
        if 'and' in topts and 'or' in topts:
            _error('Unsupported union+intersection in ', type_name, base_type)

    with open(os.path.join(__file__, '..', 'jadn_schema.json')) as f:       # Check using JSON Schema for JADN
        jsonschema.Draft7Validator(json.load(f)).validate(schema)

    with open(os.path.join(__file__, '..', 'jadn_schema.jadn')) as f:       # Optional: check using JADN meta-schema
        meta_schema = Codec(json.load(f), verbose_rec=True, verbose_str=True, config=schema)
        assert meta_schema.encode('Schema', schema) == schema

    types = set()                                                   # Additional checks not included in schema
    for t in schema['types']:
        tn = t[TypeName]
        bt = t[BaseType]
        if tn in types:
            _error('Duplicate type definition', tn)
        types |= {tn}
        if is_builtin(tn):
            _error('Reserved type name', tn)
        if not is_builtin(bt):
            _error('Invalid base type', tn + ':', bt)
        topts = topts_s2d(t[TypeOptions])
        check_typeopts(tn, bt, topts)
        flen = 0 if 'enum' in topts or 'pointer' in topts else FIELD_LENGTH[bt]
        if flen and len(t) <= Fields:
            _error('Missing fields', tn + '(' + bt + ')')
        elif not flen and len(t) > Fields:
            _error(tn + '(' + bt + ')', 'Cannot have fields')

        if flen:                # Check fields
            fids = set()        # Field IDs
            fnames = set()      # Field Names
            ordinal = bt in ('Array', 'Record')
            for n, f in enumerate(t[Fields]):
                if len(f) != flen:
                    _error('Bad field', n+1, 'in', tn, 'length', len(f), 'should be', flen)
                fids.update({f[FieldID]})
                fnames.update({f[FieldName]})
                if ordinal and f[FieldID] != n + 1:
                    _error('Item tag error:', tn + '(' + bt + ')[' + f[FieldName] + '] --', f[FieldID], 'should be', n + 1)
                if flen > FieldDesc:    # Full field, not an Enumerated item
                    fo, fto = ftopts_s2d(f[FieldOptions])
                    if is_builtin(f[FieldType]):
                        check_typeopts(tn + '/' + f[FieldName], f[FieldType], fto)
                        if 'minc' in fo and 'maxc' in fo:
                            if fo['minc'] < 0 or (fo['maxc'] > 0 and fo['maxc'] < fo['minc']):
                                _error(tn + '/', f[FieldName], 'bad cardinality', fo['minc'], fo['maxc'])
                    elif fto:
                        # _error(tn + '/' + f[FieldName] + '(' + f[FieldType] + ') cannot have Type options', fto)
                        pass    # TODO: pass 'unique' to multiplicity-based ArrayOf
            if len(t[Fields]) != len(fids) or len(t[Fields]) != len(fnames):
                _error('Duplicate field', tn, len(t[Fields]), 'fields,', len(fids), 'unique tags', len(fnames), 'unique names')
    return schema


def build_deps(schema):         # Build a Dependency dict: {TypeName: {Dep1, Dep2, ...}}
    def get_refs(tdef):         # Return all type references from a type definition
        oids = [OPTION_ID['ktype'], OPTION_ID['vtype'], OPTION_ID['and'], OPTION_ID['or']]  # Options whose value is/has a type name
        oids2 = [OPTION_ID['enum'], OPTION_ID['pointer']]                       # Options that enumerate fields
        refs = [to[1:] for to in tdef[TypeOptions] if to[0] in oids and not is_builtin(to[1:])]
        refs += [to for to in tdef[TypeOptions] if to[0] in oids2]
        if has_fields(tdef[BaseType]):
            for f in tdef[Fields]:
                if not is_builtin(f[FieldType]):
                    refs.append(f[FieldType])                               # Add reference to type name
                refs += get_refs(['', f[FieldType], f[FieldOptions], ''])   # Get refs from type opts in field (extension)
        return refs

    return {t[TypeName]: get_refs(t) for t in schema['types']}


def jadn_analyze(schema):
    def ns(name, nsids):   # Return namespace if name has a known namespace, otherwise return full name
        nsp = name.split(':')[0]
        return nsp if nsp in nsids else name

    items = build_deps(schema)
    # out, roots = topo_sort(items)
    imports = schema['meta']['imports'] if 'imports' in schema['meta'] else {}
    exports = schema['meta']['exports'] if 'exports' in schema['meta'] else []
    defs = {i for i in items} | set(imports)
    refs = set().union([ns(r, imports) for i in items for r in items[i]]) | set(exports)
    oids = (OPTION_ID['enum'], OPTION_ID['pointer'])
    refs = {r[1:] if r[0] in oids else r for r in refs}         # Reference base type for derived enums/pointers
    return {
        'unreferenced': [str(k) for k in defs - refs],
        'undefined': [str(k) for k in refs - defs],
        'cycles': [],
    }


def jadn_loads(jadn_str):
    schema = json.loads(jadn_str)
    return jadn_check(schema)


def jadn_load(fname):
    with open(fname, encoding='utf-8') as f:
        schema = json.load(f)
    return jadn_check(schema)


def jadn_dumps(schema, level=0, indent=1, strip=False, nlevel=None):
    sp = level * indent * ' '
    sp2 = (level + 1) * indent * ' '
    sep2 = ',\n' if strip else ',\n\n'
    if isinstance(schema, dict):
        sep = ',\n' if level > 0 else sep2
        lines = []
        for k in schema:
            lines.append(sp2 + '"' + k + '": ' + jadn_dumps(schema[k], level + 1, indent, strip))
        return '{\n' + sep.join(lines) + '\n' + sp + '}'
    elif isinstance(schema, list):
        sep = ',\n' if level > 1 else sep2
        vals = []
        nest = schema and isinstance(schema[0], list)       # Not an empty list
        for v in schema:
            sp3 = sp2 if nest else ''
            vals.append(sp3 + jadn_dumps(v, level + 1, indent, strip, level))
        if nest:
            spn = (nlevel if nlevel else level) * indent * ' '
            return '[\n' + sep.join(vals) + '\n' + spn + ']'
        return '[' + ', '.join(vals) + ']'
    elif isinstance(schema, (numbers.Number, type(''))):
        return json.dumps(schema)
    return '???'


def jadn_dump(schema, fname, source='', strip=False):
    with open(fname, 'w') as f:
        if source:
            f.write('"Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + '"\n\n')
        f.write(jadn_dumps(schema, strip=strip) + '\n')


def _error(*s):                     # Handle errors
    raise ValueError(*s)


class SchemaModule:
    def __init__(self, source):     # Read schema data, get module name
        self.source = None          # Filename or URL
        self.module = None          # Namespace unique name
        self.schema = None          # JADN data
        self.imports = None         # Copy of meta['imports'] or empty {}
        self.tx = None              # Type index: {type name: type definition in schema}
        self.deps = None            # Internal dependencies: {type1: {t2, t3}, type2: {t3, t4, t5}}
        self.refs = None            # External references {namespace1: {type1: {t2, t3}, ...}}
        self.used = None            # Types from this module that have been referenced {t2, t3}
        if isinstance(source, dict):    # If schema is provided, save data
            self.schema = source
        elif isinstance(source, str):   # If filename or URL is provided, load data and record source
            if '://' in source:
                pass                    # TODO: read schema from URL
            else:
                with open(source, encoding='utf-8') as f:
                    self.schema = json.load(f)
            self.source = source

        if 'meta' in self.schema:
            self.module = self.schema['meta']['module']
            self.imports = self.schema['meta']['imports'] if 'imports' in self.schema['meta'] else {}
        else:
            _error('Schema module must have a module ID')
        self.clear()

    def load(self):                 # Validate schema, build type dependencies and external references
        if not self.deps:           # Ignore if already loaded
            jadn_check(self.schema)
            self.tx = {t[TypeName]: t for t in self.schema['types']}
            self.deps = build_deps(self.schema)
            self.refs = defaultdict(lambda: defaultdict(set))
            for tn in self.deps:
                for dn in self.deps[tn].copy():     # Iterate over copy so original can be modified safely
                    if ':' in dn:
                        self.deps[tn].remove(dn)
                        nsid, typename = dn.split(':', maxsplit=1)
                        self.refs[self.imports[nsid]][tn].add(typename)

    def clear(self):
        self.used = set()

    def add_used(self, type):
        self.used.add(type)



import setuptools
with open('README.rst', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='jadn',
    version='0.5.0b4',
    author='David Kemp',
    author_email='dk190a@gmail.com',
    description='JADN schema tools',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/davaya/jadn-software',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
    python_requires='>=3.4',
    install_requires=['jsonschema'],
)